from typing import List, Dict, Optional, Union
import time


class VKExceptions:
    class ErrorUnknown(Exception):
        def __str__(self):
            return 'Произошла неизвестная ошибка. Попробуйте повторить запрос позже'

    class AppOffline(Exception):
        def __str__(self):
            return 'Приложение выключено. \n' \
                   'Необходимо включить приложение в настройках https://vk.com/editapp?id={Ваш API_ID} \n' \
                   ' или использовать тестовый режим (test_mode=1)'

    class MethodUnknown(Exception):
        def __str__(self):
            return 'Передан неизвестный метод.\n' \
                   'Проверьте, правильно ли указано название вызываемого метода: https://vk.com/dev/methods.'

    class SignUnknown(Exception):
        def __str__(self):
            return 'Неверная подпись.'

    class AuthorisationError(Exception):
        def __str__(self):
            return 'Авторизация пользователя не удалась.\n' \
                   'Убедитесь, что Вы используете верную схему авторизации.'

    class TooManyRequests(Exception):
        def __str__(self):
            return 'Слишком много запросов в секунду.\n' \
                   'Задайте больший интервал между вызовами или используйте метод execute.\n' \
                   'Подробнее об ограничениях на частоту вызовов см. на странице https://vk.com/dev/api_requests.'

    class PermissionError(Exception):
        def __str__(self):
            return 'Нет прав для выполнения этого действия.\n' \
                   'Проверьте, получены ли нужные права доступа при авторизации.\n' \
                   'Это можно сделать с помощью метода account.getAppPermissions.'

    class RequestInvalid(Exception):
        def __str__(self):
            return 'Неверный запрос.\n' \
                   'Проверьте синтаксис запроса и список используемых параметров\n' \
                   '(его можно найти на странице с описанием метода)'

    class SameTypeActions(Exception):
        def __str__(self):
            return 'Слишком много однотипных действий.\n' \
                   'Нужно сократить число однотипных обращений.\n' \
                   'Для более эффективной работы Вы можете использовать execute.'

    class ServerError(Exception):
        def __str__(self):
            return 'Произошла внутренняя ошибка сервера.\n Попробуйте повторить запрос позже.'

    class CaptchaError(Exception):
        def __str__(self):
            return 'Требуется ввод кода с картинки (Captcha). \n' \
                   'Процесс обработки этой ошибки подробно описан на отдельной странице.\n' \
                   'https://vk.com/dev/captcha_error'

    class AccessDenied(Exception):
        def __str__(self):
            return 'Доступ запрещён.\n' \
                   'Убедитесь, что Вы используете верные идентификаторы,\n' \
                   'и доступ к контенту для текущего пользователя есть в полной версии сайта.'


class ApiMethod:
    def __init__(self, access_token, v, session, calls_per_second):
        self._session = session
        self._calls_per_second = calls_per_second
        self._required_params = {
            'v': v,
            'access_token': access_token
        }

    def _call(self, method_name: str, params: dict) -> Union[dict, int, list]:
        url = 'https://api.vk.com/method/{}'.format(method_name)
        params.update(self._required_params)
        params.pop('self')
        params = {p: params[p] for p in params if params[p]}
        time_start = time.time()
        r = self._session.post(url, params=params)
        delay = 1 / (self._calls_per_second - 0.2) - (time.time() - time_start)
        if delay > 0:
            time.sleep(delay)
        try:
            r = r.json()['response']
        except Exception as e:
            print(r.json())
            raise e
        return r




import threading
from typing import List
from .VK import VK
import concurrent.futures
import math


class Parser:
    """
    Класс для удобного парсинга данных
    """

    def __init__(self, vk_accounts: List[VK] = []):
        """
        :param vk_accounts: список классов VK. Нужен для парсинга несколькими аккаунтами.
        """

        self.accounts = vk_accounts

    @staticmethod
    def form_packs(items: list, pack_size: int = 25) -> List[list]:
        """
        Преобразует список items в список списков List[List], каждый длиной pack_size или меньше

        :param items: список объектов
        :param pack_size: размер внутренних массивов

        :example:
        >>> p = Parser()
        >>> p.form_packs([1,2,3,4,5], 2)
        [[1,2], [3,4], [5]]
        """

        packs = [items[i:i+pack_size] for i in range(0, len(items), pack_size)]
        return packs

    @staticmethod
    def _parse_one_pack(vk: VK, method: str, pack: list = []):
        code = '''var i = 0;
        var items = {};
        var result = [];
        
        while ((i < 25) && (i < items.length)) {{
            result = result + {};
            i = i + 1;
        }}
        return result;
        '''.format(pack, method)

        return vk.execute(code=code)

    def parse_single_iter(self, vk: VK, method: str, packs: List[list]):
        """
        Использует один аккаунт для парсинга

        :param vk: класс VK
        :param method:
        :param packs:
        :return:

        :example:
        >>> vk = VK('access_token')
        >>> p = Parser()
        >>> friends = vk.friends.get()['items']
        >>> packs = p.form_packs(friends, 25)
        >>> results = []
        >>> for r in p.parse_single_iter(vk, "API.friends.get('user_id': items[i])", packs):
        >>>     results.append(r)
        """
        for p in packs:
            yield self._parse_one_pack(vk, method, p)

    def parse_single(self, vk: VK, method: str, packs: List[list]):
        result = []
        for r in self.parse_single_iter(vk, method, packs):
            result.extend(r)
        return result

    def process_parsing(self, vk: VK, method: str, items: list, pack_size: int = 25):
        packs = self.form_packs(items, pack_size)
        result = self.parse_single(vk, method, packs)
        return result

    def parse_threaded(self, method, items: list, pack_size: int = 25):
        items_per_account = math.ceil(len(items) / len(self.accounts))
        account_items = [items[items_per_account*i:items_per_account*(i+1)] for i in range(len(self.accounts))]

        result = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            results = [executor.submit(self.process_parsing,
                                       a,
                                       method,
                                       account_items[i],
                                       pack_size) for i, a in enumerate(self.accounts)]

        for f in concurrent.futures.as_completed(results):
            result.extend(f.result())

        return result

